



virtualenv --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
virtualenv -p "$(which python)" /tmp/venv-$$
source /tmp/venv-$$/bin/activate
python -c 'import sys ; sys.exit(not hasattr(sys, "real_prefix"))'

IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
